namespace BP
{
	#define PRIMAGEN_MODEL_FLAG_NODAMAGE 1 << 0
	#define PRIMAGEN_WOUNDED_HEAD_MODEL_PART PRIMAGEN_MODEL_PARTS
	
	const float BP_PRIMAGEN_TENDRIL_HEALTH = 40.0f;
	const float BP_PRIMAGEN_SMALLCLAW_HEALTH = 100.0f;
	const float BP_PRIMAGEN_BODY_HEALTH = 2500.0f;
	const float BP_PRIMAGEN_HEAD_HEALTH = 500.0f;

	const float BP_PRIMAGEN_TENDRIL_DMGSCALE = 2.0f;
	const float BP_PRIMAGEN_SMALLCLAW_DMGSCALE = 2.0f;
	const float BP_PRIMAGEN_BIGCLAW_DMGSCALE = 1.0f;
	const float BP_PRIMAGEN_HEAD_DMGSCALE = 2.0f;
	const float BP_PRIMAGEN_BODY_DMGSCALE = 1.0f;
	
	const array<kStr> BP_PRIMEAGEN_DESTROY_SOUNDS = { "Primagen Flinch 2.ksnd", "Primagen U Pitifull Fool.ksnd", "Primagen Suffer My Wrath.ksnd",
														"Primagen Laugh 1.ksnd", "Primagen I'll Destroy You.ksnd" };

	class Primagen : AI
	{
		bool hasSetupHealthBar;
		bool showHealthBar;
		bool showFlashEffects;
		float razorWindDamageScale;
		float explosiveDamageScale;
		float lastBodyDamage;
		int lastDamagedBoneFlags;
		int healthBarIndex;
		int healthBarRed;
		int healthBarGreen;
		int healthBarBlue;
		array<float> modelHealth(PRIMAGEN_MODEL_PARTS);
		array<float> modelMaxHealth(PRIMAGEN_MODEL_PARTS);
		array<float> modelDamageScale(PRIMAGEN_MODEL_PARTS);
		array<int> availDestroySoundIndexs;

		//------------------------------------------------------------------------------------------------------------------------
		Primagen(kActor @actor)
		{
			super(@actor);
			bugFixNoPreDamage = true;
			
			showHealthBar = false;
			hasSetupHealthBar = false;
			showFlashEffects = false;
			razorWindDamageScale = 1.0f;
			explosiveDamageScale = 1.0f;
			healthBarIndex = 0;
			healthBarRed = 255;
			healthBarGreen = 50;
			healthBarBlue = 50;
			modelDamageScale[PRIMAGEN_MODEL_PART_LEFTTENDRIL] = BP_PRIMAGEN_TENDRIL_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_MIDLEFTTENDRIL] = BP_PRIMAGEN_TENDRIL_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_RIGHTTENDRIL] = BP_PRIMAGEN_TENDRIL_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_MIDRIGHTTENDRIL] = BP_PRIMAGEN_TENDRIL_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_SMALLCLAW] = BP_PRIMAGEN_SMALLCLAW_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_BIGCLAW] = BP_PRIMAGEN_BIGCLAW_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_HEAD] = BP_PRIMAGEN_HEAD_DMGSCALE;
			modelDamageScale[PRIMAGEN_MODEL_PART_BODY] = BP_PRIMAGEN_BODY_DMGSCALE;
			modelMaxHealth[PRIMAGEN_MODEL_PART_LEFTTENDRIL] = BP_PRIMAGEN_TENDRIL_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_MIDLEFTTENDRIL] = BP_PRIMAGEN_TENDRIL_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_RIGHTTENDRIL] = BP_PRIMAGEN_TENDRIL_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_MIDRIGHTTENDRIL] = BP_PRIMAGEN_TENDRIL_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_SMALLCLAW] = BP_PRIMAGEN_SMALLCLAW_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_BIGCLAW] = BP_PRIMAGEN_BODY_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_BODY] = BP_PRIMAGEN_BODY_HEALTH;
			modelMaxHealth[PRIMAGEN_MODEL_PART_HEAD] = BP_PRIMAGEN_HEAD_HEALTH;
			for (int i = 0; i < PRIMAGEN_MODEL_PARTS; i++)
			{
				modelHealth[i] = modelMaxHealth[i];
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSerialize(kDict dict)
		{
			SERIALIZE_BOOL(showFlashEffects);
			SERIALIZE_BOOL(showHealthBar);
			SERIALIZE_FLOAT(razorWindDamageScale);
			SERIALIZE_FLOAT(explosiveDamageScale);
			SERIALIZE_INT(healthBarIndex);
			SERIALIZE_INT(healthBarRed);
			SERIALIZE_INT(healthBarGreen);
			SERIALIZE_INT(healthBarBlue);
			SERIALIZEARRAY_FLOAT(modelHealth);
			SERIALIZEARRAY_FLOAT(modelMaxHealth);
			SERIALIZEARRAY_FLOAT(modelDamageScale);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserialize(kDict& in dict)
		{
			DESERIALIZE_BOOL(showFlashEffects);
			DESERIALIZE_BOOL(showHealthBar);
			DESERIALIZE_FLOAT(razorWindDamageScale);
			DESERIALIZE_FLOAT(explosiveDamageScale);
			DESERIALIZE_INT(healthBarIndex);
			DESERIALIZE_INT(healthBarRed);
			DESERIALIZE_INT(healthBarGreen);
			DESERIALIZE_INT(healthBarBlue);
			DESERIALIZEARRAY_FLOAT(modelHealth);
			DESERIALIZEARRAY_FLOAT(modelMaxHealth);
			DESERIALIZEARRAY_FLOAT(modelDamageScale);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnLevelLoad(kDictMem@ pDict)
		{
			float smallClawHP;
			float tendrilHP;
			float bodyHP;
			float headHP;
			float dsTendril;
			float dsSmallClaw;
			float dsBigClaw;
			float dsBody;
			float dsHead;
			pDict.GetBool("ShowFlashEffects", showFlashEffects, false);
			pDict.GetBool("ShowHealthBar", showHealthBar, false);
			pDict.GetInt("HealthBarIndex", healthBarIndex, 0);
			pDict.GetInt("HealthBarRed", healthBarRed, 255);
			pDict.GetInt("HealthBarGreen", healthBarGreen, 50);
			pDict.GetInt("HealthBarBlue", healthBarBlue, 50);
			pDict.GetFloat("RazorWindDamageScale", razorWindDamageScale, 1.0f);
			pDict.GetFloat("ExplosiveDamageScale", explosiveDamageScale, 1.0f);
			pDict.GetFloat("DamageScale_Tendril", dsTendril, BP_PRIMAGEN_TENDRIL_DMGSCALE);
			pDict.GetFloat("DamageScale_SmallClaw", dsSmallClaw, BP_PRIMAGEN_SMALLCLAW_DMGSCALE);
			pDict.GetFloat("DamageScale_BigClaw", dsBigClaw, BP_PRIMAGEN_BIGCLAW_DMGSCALE);
			pDict.GetFloat("DamageScale_Head", dsHead, BP_PRIMAGEN_HEAD_DMGSCALE);
			pDict.GetFloat("DamageScale_Body", dsBody, BP_PRIMAGEN_BODY_DMGSCALE);
			pDict.GetFloat("Health_Tendril", tendrilHP, BP_PRIMAGEN_TENDRIL_HEALTH);
			pDict.GetFloat("Health_SmallClaw", smallClawHP, BP_PRIMAGEN_SMALLCLAW_HEALTH);
			pDict.GetFloat("Health_Head", headHP, BP_PRIMAGEN_HEAD_HEALTH);
			pDict.GetFloat("Health_Body", bodyHP, BP_PRIMAGEN_BODY_HEALTH);
			modelDamageScale[PRIMAGEN_MODEL_PART_LEFTTENDRIL] = dsTendril;
			modelDamageScale[PRIMAGEN_MODEL_PART_MIDLEFTTENDRIL] = dsTendril;
			modelDamageScale[PRIMAGEN_MODEL_PART_RIGHTTENDRIL] = dsTendril;
			modelDamageScale[PRIMAGEN_MODEL_PART_MIDRIGHTTENDRIL] = dsTendril;
			modelDamageScale[PRIMAGEN_MODEL_PART_SMALLCLAW] = dsSmallClaw;
			modelDamageScale[PRIMAGEN_MODEL_PART_BIGCLAW] = dsBigClaw;
			modelDamageScale[PRIMAGEN_MODEL_PART_HEAD] = dsHead;
			modelDamageScale[PRIMAGEN_MODEL_PART_BODY] = dsBody;
			modelMaxHealth[PRIMAGEN_MODEL_PART_LEFTTENDRIL] = tendrilHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_MIDLEFTTENDRIL] = tendrilHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_RIGHTTENDRIL] = tendrilHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_MIDRIGHTTENDRIL] = tendrilHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_SMALLCLAW] = smallClawHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_BIGCLAW] = bodyHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_BODY] = bodyHP;
			modelMaxHealth[PRIMAGEN_MODEL_PART_HEAD] = headHP;
			for (int i = 0; i < PRIMAGEN_MODEL_PARTS; i++)
			{
				modelHealth[i] = modelMaxHealth[i];
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnPreDamage(kDamageInfo& in dmgInfo)
		{
			self.Flags() |= AF_NOINFLICTDAMAGE;

			//only player can hurt primagen
			if (dmgInfo.source.Type() != kActor_Player)
			{
				return;
			}
			
			float damage = dmgInfo.hits;
			
			// nerf damage if razor wind
			if (!(dmgInfo.particle is null))
			{
				if(dmgInfo.particle.ParticleType() == kParticle_RazorWind)
				{
					damage *= razorWindDamageScale;
				}
			}
			
			if ((dmgInfo.flags & DF_EXPLOSIVE) != 0)
			{
				damage *= explosiveDamageScale;
			}

			// determine which part of the model was hit
			lastDamagedBoneFlags = dmgInfo.boneFlags;
			lastBodyDamage = 0.0f;
			for (int i = 0; i < PRIMAGEN_MODEL_PARTS; ++i)
			{
				int modelBit = (1 << i);
				if (i != PRIMAGEN_MODEL_PART_BODY && (lastDamagedBoneFlags & modelBit) != 0)
				{
					DoPartDamage(i, damage);
				}
			}
			DoPartDamage(PRIMAGEN_MODEL_PART_BODY, damage);
			
			if (!BP::Actor::IsDead(self) && modelHealth[PRIMAGEN_MODEL_PART_BODY] <= 0.0f && damage > 0.0f)
			{
				self.Flags() &= ~AF_NOINFLICTDAMAGE;
				self.Health() = damage;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeath(kDamageInfo& in dmgInfo)
		{
			if (showHealthBar)
			{
				Hud.DisableBar(healthBarIndex);
			}
			kActor @actor = BP::Spawn::Actor(BP_Actor_InterAnim_PrimagenDeadTired, self.Origin(), self.Yaw());
			if (actor !is null)
			{
				actor.PlaySound("sounds/shaders/Primagen Death.ksnd");
			}
			BP::Spawn::Particle("particles/zzplaceholder132.particle", @self, BP::Actor::CenterOrigin(@self, 2.5f), kQuat(self.Pitch(), 0.0f, self.Yaw()));
			self.Remove();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoPartDamage(int part, float damage)
		{
			if (modelHealth[part] <= 0.0f)
			{
				return;
			}
			
			int modelBit = (1 << part);

			if (part != PRIMAGEN_MODEL_PART_BODY)
			{
				modelHealth[part] -= damage;
				lastBodyDamage += damage * modelDamageScale[part];
			}
			else
			{
				modelHealth[part] -= damage + lastBodyDamage;
			}
			
			if (modelHealth[part] <= 0.0f)
			{
				modelHealth[part] = 0.0f;
				kVec3 fxPos = BP::Actor::CenterOrigin(@self, 1.0f);
				switch (part)
				{
					case PRIMAGEN_MODEL_PART_HEAD:
					{
						self.RenderMeshComponent().SetAttachedMeshVisibility(PRIMAGEN_WOUNDED_HEAD_MODEL_PART - 1, true);
						self.RenderMeshComponent().SetAttachedMeshVisibility(part - 1, false);
						fxPos = BP::Actor::CenterOrigin(@self, 0.9f);
						PartDestroyParticles(fxPos);
						self.PlaySound("sounds/shaders/Head Blown Off.ksnd");
						self.PlaySound("sounds/shaders/Primagen Dare You.ksnd");
						break;
					}
					case PRIMAGEN_MODEL_PART_SMALLCLAW:
						fxPos = BP::Actor::CenterOrigin(@self, 0.5f);
					case PRIMAGEN_MODEL_PART_LEFTTENDRIL:
					case PRIMAGEN_MODEL_PART_MIDLEFTTENDRIL:
					case PRIMAGEN_MODEL_PART_MIDRIGHTTENDRIL:
					case PRIMAGEN_MODEL_PART_RIGHTTENDRIL:
					{
						self.PlaySound("sounds/shaders/Head Blown Off.ksnd");
						self.PlaySound("sounds/shaders/" + BP::Array::GetUniqueElement(BP_PRIMEAGEN_DESTROY_SOUNDS, availDestroySoundIndexs));
						PartDestroyParticles(fxPos);
						self.RenderMeshComponent().SetAttachedMeshVisibility(part - 1, false);
						break;
					}
				}
			}
			else
			{
				if (showFlashEffects)
				{
					switch (part)
					{
						case PRIMAGEN_MODEL_PART_HEAD:
						case PRIMAGEN_MODEL_PART_SMALLCLAW:
						case PRIMAGEN_MODEL_PART_LEFTTENDRIL:
						case PRIMAGEN_MODEL_PART_MIDLEFTTENDRIL:
						case PRIMAGEN_MODEL_PART_MIDRIGHTTENDRIL:
						case PRIMAGEN_MODEL_PART_RIGHTTENDRIL:
						{
							StartFx("PrimagenFlashDamage", modelBit);
							break;
						}
					}
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_UserEvent(const int msg, const float x, const float y, const float z, const float f1, const float f2, const float f3)
		{
			kEnemyAIComponent @ai = self.EnemyAIComponent();
			
			ai.GetTargetInfo(true, true);
			
			if (ai.SightTarget().Target() is null)
			{
				return;
			}
			
			switch(msg)
			{
				case 1:
				{
					if (ai.SightTarget().m_fDist <= BIG_CLAW_DIST)
					{
						kDamageInfo cDamageInfo;
				
						cDamageInfo.radius = 10*256;
						cDamageInfo.flags = DF_NORMAL;
						@cDamageInfo.inflictor = @self;
						@cDamageInfo.source = @self;
						@cDamageInfo.target = @ai.SightTarget().Target();
						@cDamageInfo.particle = null;
						cDamageInfo.hits = BIG_CLAW_DAMAGE;
						
						ai.SightTarget().Target().InflictDamage(cDamageInfo);
					}
					break;
				}
				case 2:
				{
					if (ai.SightTarget().m_fDist <= SMALL_CLAW_DIST && modelHealth[PRIMAGEN_MODEL_PART_SMALLCLAW] > 0.0f)
					{
						kDamageInfo cDamageInfo;
				
						cDamageInfo.radius = 10*256;
						cDamageInfo.flags = DF_NORMAL;
						@cDamageInfo.inflictor = @self;
						@cDamageInfo.source = @self;
						@cDamageInfo.target = @ai.SightTarget().Target();
						@cDamageInfo.particle = null;
						cDamageInfo.hits = SMALL_CLAW_DAMAGE;
						
						ai.SightTarget().Target().InflictDamage(cDamageInfo);
					}
					break;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void PartDestroyParticles(const kVec3&in pos)
		{
			kQuat rotation(self.Pitch(), 0.0f, self.Yaw());
			BP::Spawn::Particle("particles/Hot_Point_Particle_13.particle", @self, pos, rotation);
			BP::Spawn::Particle("particles/Hot_Point_Particle_14.particle", @self, pos, rotation);
			BP::Spawn::Particle("particles/zPlaceholder_12.particle", @self, pos, rotation);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void StartFx(const kStr&in szFxFile, const int modelBits)
		{
			int bits = 0;
			for (int i = 1; i < PRIMAGEN_MODEL_PARTS; ++i)
			{
				if (((1 << i) & modelBits) != 0)
				{
					bits |= (1 << (i-1));
				}
			}
			self.RenderMeshComponent().FxFlags() = bits;
			self.RunFxEvent(szFxFile);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnTick()
		{
			if (showHealthBar)
			{
				if (!hasSetupHealthBar)
				{
					//if sighttarget or attack target is player then show the health bar if hasn't been shown yet
					kEnemyAIComponent @ai = self.EnemyAIComponent();
					ai.GetTargetInfo(true, true);
					kActor @target = ai.SightTarget().Target();
					if (target !is null && target.Type() == kActor_Player)
					{
						hasSetupHealthBar = true;
						Hud.SetBar(healthBarIndex, healthBarRed, healthBarGreen, healthBarBlue, 255, 100);
					}
				}
				
				float currentHealth = modelHealth[PRIMAGEN_MODEL_PART_BODY];
				float maxHealth = modelMaxHealth[PRIMAGEN_MODEL_PART_BODY];
				if (maxHealth > 0.0f)
				{
					int healthPercent = int((currentHealth / maxHealth) * 100.0f);
					Hud.UpdateBar(healthBarIndex, MAX(healthPercent, 0));
				}
				else
				{
					Hud.UpdateBar(healthBarIndex, 0);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
