#define QUEEN_COMBAT1_DIST				500.0
#define QUEEN_COMBAT1_DAMAGE			8.0
#define QUEEN_COMBAT1_ANIMHITTIME	0.5
#define QUEEN_COMBAT1_WAITTIME		1.0

namespace BP
{
	class Queen : AI
	{
		bool showHealthBar;
		float razorWindDamageScale;
		float explosiveDamageScale;
		int healthBarIndex;
		int healthBarRed;
		int healthBarGreen;
		int healthBarBlue;
		float maxHealth;
		
		bool hasSetupHealthBar;
		float combat1WaitTime;
		kEnemyAIComponent@ ai;
		kAnimTrackComponent@ anim;
		//------------------------------------------------------------------------------------------------------------------------
		Queen(kActor @actor)
		{
			super(@actor);
			bugFixNoPreDamage = true;
			
			showHealthBar = false;
			razorWindDamageScale = 0.25f;
			explosiveDamageScale = 0.5f;
			healthBarIndex = 0;
			healthBarRed = 255;
			healthBarGreen = 50;
			healthBarBlue = 50;
			maxHealth = 1000.0f;
			hasSetupHealthBar = false;
			combat1WaitTime = 0.0f;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSerialize(kDict dict)
		{
			AI::BP_OnSerialize(dict);
			SERIALIZE_FLOAT(combat1WaitTime);
			SERIALIZE_FLOAT(maxHealth);
			SERIALIZE_BOOL(showHealthBar);
			SERIALIZE_INT(healthBarIndex);
			SERIALIZE_INT(healthBarRed);
			SERIALIZE_INT(healthBarGreen);
			SERIALIZE_INT(healthBarBlue);
			SERIALIZE_FLOAT(razorWindDamageScale);
			SERIALIZE_FLOAT(explosiveDamageScale);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserialize(kDict& in dict)
		{
			AI::BP_OnDeserialize(dict);
			DESERIALIZE_FLOAT(combat1WaitTime);
			DESERIALIZE_FLOAT(maxHealth);
			DESERIALIZE_BOOL(showHealthBar);
			DESERIALIZE_INT(healthBarIndex);
			DESERIALIZE_INT(healthBarRed);
			DESERIALIZE_INT(healthBarGreen);
			DESERIALIZE_INT(healthBarBlue);
			DESERIALIZE_FLOAT(razorWindDamageScale);
			DESERIALIZE_FLOAT(explosiveDamageScale);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnLevelLoad(kDictMem@ pDict) override
		{
			AI::BP_OnLevelLoad(pDict);
			pDict.GetBool("ShowHealthBar", showHealthBar, false);
			pDict.GetInt("HealthBarIndex", healthBarIndex, 0);
			pDict.GetInt("HealthBarRed", healthBarRed, 255);
			pDict.GetInt("HealthBarGreen", healthBarGreen, 50);
			pDict.GetInt("HealthBarBlue", healthBarBlue, 50);
			pDict.GetFloat("RazorWindDamageScale", razorWindDamageScale, 0.25f);
			pDict.GetFloat("ExplosiveDamageScale", explosiveDamageScale, 0.5f);
			maxHealth = self.Health();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) override
		{
			self.Flags() |= AF_NOINFLICTDAMAGE;
			
			float damage = dmgInfo.hits;
			
			//Damage modifiers
			if (dmgInfo.particle !is null && dmgInfo.particle.ParticleType() == kParticle_RazorWind)
			{
				damage *= razorWindDamageScale;
			}
			if ((dmgInfo.flags & DF_EXPLOSIVE) != 0)
			{
				damage *= explosiveDamageScale;
			}
			
			//inflict new damage
			if (!BP::Actor::IsDead(self) && damage > 0.0f)
			{
				self.Health() -= damage;
				if (self.Health() <= 0.0f)
				{
					self.Flags() &= ~AF_NOINFLICTDAMAGE;
					self.Health() = damage;
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeath(kDamageInfo& in dmgInfo) override
		{
			AI::BP_OnDeath(dmgInfo);
			if (showHealthBar)
			{
				Hud.DisableBar(healthBarIndex);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		// Checks for combat1 animation in order to apply damage to target
		//------------------------------------------------------------------------------------------------------------------------
		void CheckCombat1Radius()
		{
			combat1WaitTime -= GAME_DELTA_TIME;
			
			if (anim.IsPlaying(ANIM_GROUND_ATTACK_COMBAT1) && anim.PlayTime() >= QUEEN_COMBAT1_ANIMHITTIME && combat1WaitTime <= 0.0f)
			{
				combat1WaitTime = QUEEN_COMBAT1_WAITTIME;
				
				ai.GetTargetInfo(true, true);
				if (ai.SightTarget().Target() is null)
				{
					return;
				}
				
				if (ai.SightTarget().m_fDist <= QUEEN_COMBAT1_DIST)
				{
					kDamageInfo cDamageInfo;
					cDamageInfo.radius = 10*256;
					cDamageInfo.flags = DF_NORMAL;
					@cDamageInfo.inflictor = @self;
					@cDamageInfo.source = @self;
					@cDamageInfo.target = @ai.SightTarget().Target();
					@cDamageInfo.particle = null;
					cDamageInfo.hits = QUEEN_COMBAT1_DAMAGE;
					ai.SightTarget().Target().InflictDamage(cDamageInfo);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnStart() override
		{
			AI::BP_OnStart();
			@ai = self.EnemyAIComponent();
			@anim = self.AnimTrackComponent();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserializeStart() override
		{
			AI::BP_OnDeserializeStart();
			@ai = self.EnemyAIComponent();
			@anim = self.AnimTrackComponent();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnTick() override
		{
			AI::BP_OnTick();
			CheckCombat1Radius();
			
			if (showHealthBar)
			{
				if (!hasSetupHealthBar)
				{
					//if sighttarget or attack target is player then show the health bar if hasn't been shown yet
					ai.GetTargetInfo(true, true);
					kActor @target = ai.SightTarget().Target();
					if (target !is null && target.Type() == kActor_Player)
					{
						hasSetupHealthBar = true;
						Hud.SetBar(healthBarIndex, healthBarRed, healthBarGreen, healthBarBlue, 255, 100);
					}
				}
				
				if (maxHealth > 0.0f)
				{
					int healthPercent = int((self.Health() / maxHealth) * 100.0f);
					Hud.UpdateBar(healthBarIndex, MAX(healthPercent, 0));
				}
				else
				{
					Hud.UpdateBar(healthBarIndex, 0);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
