namespace BP
{	
	//mod settings
	// float playerGravity = 0.6f; //value for player gravity
	// float playerSpeed = 0.5f; //value for friction and air friction
	// float playerReceiveDamage = 1.0f; //damege scale the player will Receive
	// bool playerCanDoubleJump = false;
	// bool playerCanGroundDash = false;
	// bool playerCanAirDash = false;
	// bool playerCanGroundPound = false;
	// bool showPlayerDamage = false;
	// bool playerCloaking = false;
	// float playerDamage = 1.0f; //the amount of damage players do on ai and players
	// kVec3 slapVelocity;
	// int mpGameMode = MP_GAMEMODE_ROK;
	
	const float PLAYER_AIR_JUMP_VELOCITY = 10.0f;
	const float PLAYER_GROUND_POUND_HEIGHT = 300.0f;
	const float PLAYER_GROUND_POUND_VELOCITY = -60.0f;
	
	const float PLAYER_DASH_LENGTH = 0.4f;
	const float PLAYER_DASH_INPUT_TIME = 0.25f;
	const float PLAYER_DASH_COOLDOWN = 0.5f;

	class Player : BP::ScriptActor
	{
		int airJumpsDone;
		bool wasDead;
		bool wasUnderWater;
		bool wasGrounded;
		bool airDashEnabled;
		bool groundPound;
		float dashInputTime; //double tap dash time (time between inputs)
		float dashCooldown; //time until can dash again
		float dashTimeLength; //How long the dash effect last has left
		float dashForwardSpeed = 8.0f;
		float dashJumpSpeed = 5.0f;
		float dashForwardDirectionX;
		float dashForwardDirectionY;
		float dashLastYawOffset;
		uint16 lastButtons;
		bool hasMPActiveTicked; //has the player ticked in MP while being active by a player
		
		//-----used because of onPreDamage, OnDamage, and OnDeath get called twice bug (Remove when patched)-----
		// bool ignoreNextOnPreDamage = true;
		// bool ignoreNextOnDamage = true;
		// bool ignoreNextOnDeath = true;
		
		//------------------------------------------------------------------------------------------------------------------------
		Player(kActor @actor)
		{
			super(@actor);
			BP::Map::OnCreate();
			// BP::Spawn::Actor(BP_Actor_AI_PrimagenBoss, kVec3(0,1000,0));
			// BP::Spawn::Actor(BP_Actor_AI_PrimagenBoss, kVec3(200,1000,0));
			// BP::Spawn::Actor(BP_Actor_AI_PrimagenBoss, kVec3(-200,1000,0));
			// BP::Spawn::Actor(BP_Actor_AI_PrimagenBoss, kVec3(0,1200,0));
			// BP::Spawn::Actor(BP_Actor_AI_PrimagenBoss, kVec3(0,1400,0));
		}
		//------------------------------------------------------------------------------------------------------------------------
		~Player()
		{
			BP::Map::OnDestroy();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSerialize(kDict dict) override
		{
			SERIALIZE_FLOAT(dashInputTime);
			SERIALIZE_FLOAT(dashCooldown);
			SERIALIZE_FLOAT(dashTimeLength);
			SERIALIZE_FLOAT(dashForwardSpeed);
			SERIALIZE_FLOAT(dashJumpSpeed);
			SERIALIZE_FLOAT(dashForwardDirectionX);
			SERIALIZE_FLOAT(dashForwardDirectionY);
			
			SERIALIZE_INT(airJumpsDone);
			SERIALIZE_BOOL(wasGrounded);
			SERIALIZE_BOOL(airDashEnabled);
			SERIALIZE_BOOL(wasUnderWater);
			SERIALIZE_BOOL(groundPound);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnDeserialize(kDict& in dict) override
		{
			DESERIALIZE_FLOAT(dashInputTime);
			DESERIALIZE_FLOAT(dashCooldown);
			DESERIALIZE_FLOAT(dashTimeLength);
			DESERIALIZE_FLOAT(dashForwardSpeed);
			DESERIALIZE_FLOAT(dashJumpSpeed);
			DESERIALIZE_FLOAT(dashForwardDirectionX);
			DESERIALIZE_FLOAT(dashForwardDirectionY);

			DESERIALIZE_INT(airJumpsDone);
			DESERIALIZE_BOOL(wasGrounded);
			DESERIALIZE_BOOL(airDashEnabled);
			DESERIALIZE_BOOL(wasUnderWater);
			DESERIALIZE_BOOL(groundPound);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) override
		{
			if (IsMultiplayer())
			{
				//for on damage twice bug fix ignore every other on damage
				//(When bug is patched remove the ignoreNextOnPreDamage var
				// ignoreNextOnPreDamage = !ignoreNextOnPreDamage;
				// if (!ignoreNextOnPreDamage && !BP::Actor::IsDead(@self))
				if (!BP::Actor::IsDead(@self))
				{
					kActor @src = dmgInfo.source;
					float realDamage = dmgInfo.hits;
					float moreHits = realDamage * BP::playerReceiveDamage;
					if (src !is null && src.Type() == kActor_Player)
					{
						moreHits *= BP::playerDamage;
					}
					//increase the razorwind damage output in multiplayer for some reason it's really low for players only
					kParticle @particle = dmgInfo.particle;
					if (particle !is null && particle.ParticleType() == kParticle_RazorWind)
					{
						moreHits *= 500.0f;
					}
					float hitsDif = moreHits - realDamage;
					self.Health() = MAX(self.Health() - hitsDif, realDamage);
					ShowDamage(@src, moreHits);
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TryAirJump()
		{
			//can't double jump or is climbing, or in water
			if (!playerCanDoubleJump || (self.WorldComponent().Flags() & (WCF_AT_CLIMBING_SURFACE | WCF_UNDER_WATER | WCF_ON_WATER)) != 0)
			{
				return;
			}
			
			kPuppet @puppet = self.CastToPuppet();
			bool doingJump = (puppet.PlayerFlags() & PF_JUMPING) != 0;
			bool hasJumped = (puppet.PlayerFlags() & PF_HASJUMPED) != 0;
			if (!BP::Actor::IsGrounded(@self) && !doingJump && hasJumped && airJumpsDone < 1) //BP::Game::playerMaxAirJumps
			{
				kVec3 vel = self.MovementComponent().Velocity();
				self.MovementComponent().Velocity() = kVec3(vel.x, vel.y, PLAYER_AIR_JUMP_VELOCITY);
				airJumpsDone++;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TryGroundPound()
		{
			//can't groundpound or is climbing, or in water
			if (!playerCanGroundPound || (self.WorldComponent().Flags() & (WCF_AT_CLIMBING_SURFACE | WCF_UNDER_WATER | WCF_ON_WATER)) != 0)
			{
				return;
			}
			
			if (!groundPound && !BP::Actor::IsGrounded(self) && self.Origin().z >= self.WorldComponent().FloorHeight() + PLAYER_GROUND_POUND_HEIGHT)
			{
				groundPound = true;
				kPuppet @puppet = self.CastToPuppet();
				puppet.PlayerFlags() |= PF_HASJUMPED;
				airDashEnabled = false;
				self.MovementComponent().Velocity() = kVec3(0.0f, 0.0f, PLAYER_GROUND_POUND_VELOCITY);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TickDash()
		{
			dashInputTime -= GAME_DELTA_TIME;
			dashCooldown -= GAME_DELTA_TIME;
			dashTimeLength -= GAME_DELTA_TIME;
			if (dashTimeLength > 0.0f)
			{
				float speed = Math::Sin((dashTimeLength / PLAYER_DASH_LENGTH) * Math::pi) * dashForwardSpeed;
				self.MovementComponent().Velocity() += kVec3(dashForwardDirectionX * speed, dashForwardDirectionY * speed, 0.0f);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		// if is able to dash then dashs in the forward direction + yawOffset of the actor. Returns true if did dash.
		//------------------------------------------------------------------------------------------------------------------------
		bool TryDashInput(float yawOffset)
		{
			bool didDash = false;
			//last input was in the same direction as current input and isn't on cooldown and pressed within a certain time
			if (Math::Approximately(dashLastYawOffset, yawOffset) && dashCooldown <= 0.0f && dashInputTime > 0.0f && dashInputTime < PLAYER_DASH_INPUT_TIME)
			{
				dashInputTime = 0.0f;
				dashCooldown = PLAYER_DASH_COOLDOWN;
				dashTimeLength = PLAYER_DASH_LENGTH;
				self.MovementComponent().Velocity().z += dashJumpSpeed;
				dashForwardDirectionX = Math::Sin(self.Yaw() + yawOffset);
				dashForwardDirectionY = Math::Cos(self.Yaw() + yawOffset);
				didDash = true;
			}
			else
			{
				dashInputTime = PLAYER_DASH_INPUT_TIME;
			}
			
			dashLastYawOffset = yawOffset;
			return didDash;
		}
		//------------------------------------------------------------------------------------------------------------------------
		void TryDash(float yawOffset)
		{			
			//climbing, or in water
			if ((self.WorldComponent().Flags() & (WCF_AT_CLIMBING_SURFACE | WCF_UNDER_WATER | WCF_ON_WATER)) != 0)
			{
				return;
			}
				
			if (BP::Actor::IsGrounded(self))
			{
				if (!playerCanGroundDash)
				{
					return;
				}
				
				if (dashCooldown <= 0.0f)
				{
					bool didGroundDash = TryDashInput(yawOffset);
					if (didGroundDash)
					{
						self.PlaySound(400); //Talon Swoosh
					}
				}
			}
			else
			{
				if (!playerCanAirDash)
				{
					return;
				}
				
				//airDash allowed and dash not on cooldown
				if (airDashEnabled && dashCooldown <= 0.0f)
				{
					bool didAirDash = TryDashInput(yawOffset);
					if (didAirDash)
					{
						
						self.PlaySound(400); //Talon Swoosh
						airDashEnabled = false;
					}
				}
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnGroundedChanged(bool isGrounded)
		{
			if (isGrounded)
			{
				if (groundPound)
				{
					groundPound = false;
					
					BP::Spawn::Particle(BP_Particle_GroundPound, @self, self.Origin() + kVec3(0.0f, 0.0f, 20.0f));
					SpawnShake(kVec3(0, 0, 125.0f), self.Origin(), 10000, true);
				}
				
				airJumpsDone = 0;
				airDashEnabled = true;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnUnderWaterChanged(bool isUnderWater)
		{
			if (groundPound)
			{
				groundPound = false;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void OnDeadChanged(bool isDead)
		{
			if (!isDead)
			{
				BP::UserEvent::PlayerRevived(@BP::map, @self);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void SetPlayerModStats()
		{
			BP::Actor::SetGravity(@self, MAX(BP::playerGravity, 0.0f));
			float speed = Math::Clampf(BP::playerSpeed, 0.0f, 1.0f);
			BP::Actor::SetSpeed(@self, speed);
			if (BP::playerCloaking)
			{
				self.RenderMeshComponent().Flags() |= RMCF_CLOAKED;
			}
			else
			{
				self.RenderMeshComponent().Flags() &= ~RMCF_CLOAKED;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		// void BP_OnSpawn()
		// {
		// }
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnStart() override
		{
			BP::UserEvent::PlayerStarted(@BP::map, @self);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnTick() override
		{
			bool isDead = BP::Actor::IsDead(@self);
			if (wasDead != isDead)
			{
				OnDeadChanged(isDead);
				wasDead = isDead;
			}

			kPuppet @puppet = self.CastToPuppet();
			kPlayer player = puppet.PlayerOwner();
			TickDash();
			SetPlayerModStats();

			if (IsMultiplayer())
			{
				if (!hasMPActiveTicked)
				{
					hasMPActiveTicked = true;
					// SetPlayerModStats();
				}
				
				if (BP::Actor::IsMPPlayerNotInGame(@self))
				{
					//don't tick player actors that are not being controlled by a user
					return;
				}
			}
						
			bool isUnderWater = (self.WorldComponent().Flags() & WCF_UNDER_WATER) != 0;
			if (wasUnderWater != isUnderWater)
			{
				OnUnderWaterChanged(isUnderWater);
				wasUnderWater = isUnderWater;
			}
			
			bool isGrounded = BP::Actor::IsGrounded(@self);
			if (wasGrounded != isGrounded)
			{
				OnGroundedChanged(isGrounded);
				wasGrounded = isGrounded;
			}
			
			if (groundPound && self.MovementComponent().Velocity().z >= 0.0f)
			{
				groundPound = false;
			}
			
			if (isDead)
			{
				return;
			}
			
			//Tick Input
			uint16 buttons = player.Buttons();
			
			if (BP::Input::ForwardDown(buttons, lastButtons))
			{
				TryDash(0.0f);
			}
			else if (BP::Input::BackwardDown(buttons, lastButtons))
			{
				TryDash(Math::pi);
			}
			
			if (BP::Input::StrafeLeftDown(buttons, lastButtons))
			{
				TryDash(-Math::piHalf);
			}
			else if (BP::Input::StrafeRightDown(buttons, lastButtons))
			{
				TryDash(Math::piHalf);
			}
			
			if (BP::Input::JumpDown(buttons, lastButtons))
			{
				TryAirJump();
			}
			else if (BP::Input::CrouchDown(buttons, lastButtons))
			{
				TryGroundPound();
			}
			
			// if (BP::Input::ScopeZoomInDown(buttons, lastButtons))
			// {

			// }
			// if (BP::Input::ScopeZoomOutDown(buttons, lastButtons))
			// {

			// }
			
			lastButtons = buttons;
		}
		//------------------------------------------------------------------------------------------------------------------------
	};
}
