namespace BP
{
	int lastInstanceID;
	abstract class ScriptActor : ScriptObject
	{
		kActor@ self;
		int instanceID;
		bool hasTicked;
		bool hasDeserialized;
		bool didOnSpawn;
		float scheduledNetSyncTime;
		bool netSendOnPlayerJoin = false;
		
		bool ignoreNextOnPreDamage = true;
		bool bugFixNoPreDamage = false;
		bool ignoreNextOnDamage = true;
		bool ignoreNextOnDeath = true;
		//------------------------------------------------------------------------------------------------------------------------
		ScriptActor(kActor@ actor)
		{
			@self = actor;
			
			lastInstanceID += (lastInstanceID == -1) ? 2 : 1;
			instanceID = lastInstanceID;
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnSpawn() final
		{
			BP_OnSpawn();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnBeginLevel() final
		{
			BP_OnBeginLevel();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnLevelLoad(kDictMem@ pDict) final
		{
			didOnSpawn = true;
			BP_OnLevelLoad(@pDict);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnSerialize(kDict& out dict) final
		{
			SERIALIZE_INT(instanceID);
			SERIALIZE_BOOL(didOnSpawn);
			SERIALIZE_BOOL(bugFixNoPreDamage);
			BP_OnSerialize(dict);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnDeserialize(kDict& in dict) final
		{
			DESERIALIZE_INT(instanceID);
			DESERIALIZE_BOOL(didOnSpawn);
			DESERIALIZE_BOOL(bugFixNoPreDamage);
			BP_OnDeserialize(dict);
			hasDeserialized = true;
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnTrigger(kActor@ pInstigator, const int msg) final
		{
			BP_OnTrigger(@pInstigator, msg);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnTick() final
		{
			if (!hasTicked)
			{
				if (self.GameTicks() == 0)
				{
					BP_OnStart();
				}
				if (hasDeserialized)
				{
					BP_OnDeserializeStart();
				}
				hasTicked = true;
			}
			
			if (IsNetServer && scheduledNetSyncTime > 0.0f)
			{
				scheduledNetSyncTime -= GAME_DELTA_TIME;
				if (scheduledNetSyncTime <= 0.0f)
				{
					BP_OnNetSend();
				}
			}
			BP_OnTick();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ScheduleNetSync()
		{
			if (scheduledNetSyncTime <= 0.0f)
			{
				scheduledNetSyncTime = NET_SYNC_TIME;
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnEndLevel() final
		{
			BP_OnEndLevel();
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnPreDamage(kDamageInfo& in dmgInfo) final
		{
			if (!didOnSpawn)
			{
				ignoreNextOnPreDamage = !ignoreNextOnPreDamage;
				if (ignoreNextOnPreDamage)
				{
					if (bugFixNoPreDamage)
					{
						self.Flags() |= AF_NOINFLICTDAMAGE;
					}
					return;
				}
			}
			BP_OnPreDamage(dmgInfo);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnDamage(kDamageInfo& in dmgInfo) final
		{
			if (!didOnSpawn)
			{
				ignoreNextOnDamage = !ignoreNextOnDamage;
				if (ignoreNextOnDamage)
				{
					return;
				}
			}
			BP_OnDamage(dmgInfo);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnDeath(kDamageInfo& in dmgInfo) final
		{
			if (!didOnSpawn)
			{
				ignoreNextOnDeath = !ignoreNextOnDeath;
				if (ignoreNextOnDeath)
				{
					return;
				}
			}
			BP_OnDeath(dmgInfo);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnTouch(kActor@ pInstigator) final
		{
			BP_OnTouch(@pInstigator);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void OnCollide(kActor@ pCollider) final
		{
			BP_OnCollide(@pCollider);
		}
		//------------------------------------------------------------------------------------------------------------------------
		private void UserEvent(const float x, const float y, const float z, const float f1, const float f2, const float f3, const float f4) final
		{
			int msg = int(f1);
			
			switch (msg)
			{
				case USEREVENT_MSG_GET_SCRIPT:
				{
					@BP::UserEvent::scriptResult = @this;
					break;
				}
				case USEREVENT_MSG_MP_PLAYER_JOIN:
				{
					if (netSendOnPlayerJoin)
					{
						DoNetSend();
					}
					break;
				}
				case USEREVENT_MSG_ACTOR_NETSEND:
				{
					DoNetSend();
					break;
				}
				case USEREVENT_MSG_ACTOR_NETRECEIVE:
				{
					BP_OnNetReceive(BP::UserEvent::argInts, BP::UserEvent::argShorts);
					break;
				}
				
			}

			BP_UserEvent(msg, x, y, z, f2, f3, f4);
		}
		//------------------------------------------------------------------------------------------------------------------------
		void DoNetSend()
		{
			scheduledNetSyncTime = 0.0f;
			BP_OnNetSend();
		}
		//------------------------------------------------------------------------------------------------------------------------
		void ShowDamage(kActor @src, const float damage)
		{
			if (BP::showPlayerDamage && IsMultiplayer() && src !is null)
			{
				int damageDisplay = (damage > 0.0f && damage < 1.0f) ? Math::CeilToInt(damage) : int(damage);
				BP::Actor::ShowWorldText(@src, @self, "" + damageDisplay);
			}
		}
		//------------------------------------------------------------------------------------------------------------------------
		void BP_OnSpawn() {}
		void BP_OnBeginLevel() {}
		void BP_OnEndLevel() {}
		void BP_OnLevelLoad(kDictMem@ pDict) {}
		void BP_OnSerialize(kDict dict) {}
		void BP_OnDeserialize(kDict& in dict) {}
		void BP_OnTrigger(kActor@ pInstigator, const int msg) {}
		void BP_OnTick() {}
		void BP_OnStart() {} //called once before tick and never again even after loading a save game
		void BP_OnDeserializeStart() {} //called once before tick after loading a save game
		void BP_OnPreDamage(kDamageInfo& in dmgInfo) {}
		void BP_OnDamage(kDamageInfo& in dmgInfo) {}
		void BP_OnDeath(kDamageInfo& in dmgInfo) {}
		void BP_OnTouch(kActor@ pInstigator) {}
		void BP_OnCollide(kActor@ pCollider) {}
		void BP_UserEvent(const int msg, const float x, const float y, const float z, const float f1, const float f2, const float f3) {}
		void BP_OnNetSend() {} // called when need to sync with players
		void BP_OnNetReceive(array<int>& int32Args, array<int16>& int16Args) {} //called when receiving Network data
		//------------------------------------------------------------------------------------------------------------------------
	};
}
