#include "scripts/BP/Utils/Debug.cpp"

#define PrintLog(x)				Sys.Print("" + x)
#define PrintError(x)			Sys.Warning("" + x + "\n")

#define SERIALIZE_INT(x)		dict.Add(#x, ""+x)
#define SERIALIZE_FLOAT(x)		dict.Add(#x, ""+x)
#define SERIALIZE_BOOL(x)		dict.Add(#x, ""+x)
#define SERIALIZE_STRING(x)		dict.Add(#x, x)
#define SERIALIZE_VECTOR(x)		dict.Add(#x, x.ToString())
#define DESERIALIZE_STRING(x)	dict.GetString(#x, x)
#define DESERIALIZE_VECTOR(x)	dict.GetVector(#x, x)

#define SERIALIZEKEY_INT(key, x)		dict.Add(key + #x, ""+x)
#define SERIALIZEKEY_FLOAT(key, x)		dict.Add(key + #x, ""+x)
#define SERIALIZEKEY_BOOL(key, x)		dict.Add(key + #x, ""+x)
#define SERIALIZEKEY_STRING(key, x)		dict.Add(key + #x, x)
#define SERIALIZEKEY_VECTOR(key, x)		dict.Add(key + #x, x.ToString())
#define DESERIALIZEKEY_FLOAT(key, x)	dict.GetFloat(key + #x, x)
#define DESERIALIZEKEY_INT(key, x)		dict.GetInt(key + #x, x)
#define DESERIALIZEKEY_BOOL(key, x)		x = DeserializeBool(dict, key + #x)
#define DESERIALIZEKEY_STRING(key, x)	dict.GetString(key + #x, x)
#define DESERIALIZEKEY_VECTOR(key, x)	dict.GetVector(key + #x, x)

#define SERIALIZEARRAY_INT(ary)			SerializeArrayInt(dict, #ary, ary)
#define DESERIALIZEARRAY_INT(ary)		DeserializeArrayInt(dict, #ary, ary)
#define SERIALIZEARRAY_FLOAT(ary)		SerializeArrayFloat(dict, #ary, ary)
#define DESERIALIZEARRAY_FLOAT(ary)		DeserializeArrayFloat(dict, #ary, ary)

void SerializeArrayInt(kDict& in dict, kStr& in key, const array<int>& ary)
{
	int length = int(ary.length());
	SERIALIZEKEY_INT(key + "Length", length);
	for (int i = 0; i < length; i++)
	{
		SERIALIZEKEY_INT(key + i, ary[i]);
	}
}
			
void DeserializeArrayInt(kDict& in dict, kStr& in key, array<int>& ary)
{
	int length;
	DESERIALIZEKEY_INT(key + "Length", length);
	ary.resize(length);
	for (int i = 0; i < length; i++)
	{
		DESERIALIZEKEY_INT(key + i, ary[i]);
	}
}

void SerializeArrayFloat(kDict& in dict, kStr& in key, const array<float>& ary)
{
	int length = int(ary.length());
	SERIALIZEKEY_INT(key + "Length", length);
	for (int i = 0; i < length; i++)
	{
		SERIALIZEKEY_FLOAT(key + i, ary[i]);
	}
}
			
void DeserializeArrayFloat(kDict& in dict, kStr& in key, array<float>& ary)
{
	int length;
	DESERIALIZEKEY_INT(key + "Length", length);
	ary.resize(length);
	for (int i = 0; i < length; i++)
	{
		DESERIALIZEKEY_FLOAT(key + i, ary[i]);
	}
}

#define QUAT_IDENTITY kQuat(0.0f, 0.0f, 0.0f)
#define VEC_ZERO Math::vecZero

#define FOREACH_ACTOR \
	kActorIterator iter; \
	kActor@ actor; \
	while ((@actor = iter.GetNext()) !is null)
		
#define IsNetServer (Network.IsNetgame() && Network.IsServer())
#define IsNetClient (Network.IsNetgame() && Network.IsClient())

#define NET_SYNC_TIME 0.033333f

#define USEREVENT_MSG_SPAWN					100
#define USEREVENT_MSG_GET_SCRIPT			101
#define USEREVENT_MSG_AI_MOVE_POINT			102
#define USEREVENT_MSG_AI_ATTACK_POINT		103
#define USEREVENT_MSG_AI_STARTED			104
#define USEREVENT_MSG_AI_KILLED				105
#define USEREVENT_MSG_PLAYER_REVIVED        106
#define USEREVENT_MSG_PLAYER_STARTED        107
#define USEREVENT_MSG_MP_PLAYER_JOIN		108
#define USEREVENT_MSG_ACTOR_NETSEND			109
#define USEREVENT_MSG_ACTOR_NETRECEIVE		110

#define MP_GAMEMODE_ROK				1
#define MP_GAMEMODE_TEAMROK			2
#define MP_GAMEMODE_FRAGTAG			3
#define MP_GAMEMODE_LASTSTANDING	4
#define MP_GAMEMODE_ROKMIX			5
#define MP_GAMEMODE_INSTAROK		6
#define MP_GAMEMODE_RAPTORFEST		7

#include "defs/BP/common.cpp"

#include "scripts/BP/Utils/Math.cpp"
#include "scripts/BP/Utils/String.cpp"
#include "scripts/BP/Utils/Array.cpp"
#include "scripts/BP/Utils/Actor.cpp"
#include "scripts/BP/Utils/Input.cpp"
#include "scripts/BP/Utils/Spawn.cpp"
#include "scripts/BP/Utils/UserEvent.cpp"
#include "scripts/BP/Utils/Inventory.cpp"
#include "scripts/BP/Utils/Cvars.cpp"
#include "scripts/BP/Utils/Map.cpp"
#include "scripts/BP/Utils/Words.cpp"
#include "scripts/BP/Utils/Hud.cpp"
#include "scripts/BP/Utils/Game.cpp"

#include "scripts/BP/Actors/ScriptActor.cpp"
#include "scripts/BP/Actors/Player.cpp"
#include "scripts/BP/Actors/MoveablePlatform.cpp"
#include "scripts/BP/Actors/MusicWav.cpp"
#include "scripts/BP/Actors/DifficultyCheck.cpp"
#include "scripts/BP/Actors/Destructible.cpp"
#include "scripts/BP/Actors/Action.cpp"
#include "scripts/BP/Actors/InteractiveAnim.cpp"
#include "scripts/BP/Actors/Door.cpp"

#include "scripts/BP/Actors/AI/AI.cpp"
#include "scripts/BP/Actors/AI/Primagen.cpp"
#include "scripts/BP/Actors/AI/Queen.cpp"
#include "scripts/BP/Actors/AI/AraissiSoldier.cpp"
#include "scripts/BP/Actors/AI/BioBot.cpp"
#include "scripts/BP/Actors/AI/BlindGuardian.cpp"
#include "scripts/BP/Actors/AI/BlindSentinel.cpp"
#include "scripts/BP/Actors/AI/CaveSpider.cpp"
#include "scripts/BP/Actors/AI/CaveWorm.cpp"
#include "scripts/BP/Actors/AI/Compsognathus.cpp"
#include "scripts/BP/Actors/AI/Deadman.cpp"
#include "scripts/BP/Actors/AI/DeadmanHalf.cpp"
#include "scripts/BP/Actors/AI/DeathGuard.cpp"
#include "scripts/BP/Actors/AI/Endtrail.cpp"
#include "scripts/BP/Actors/AI/Fireborn.cpp"
#include "scripts/BP/Actors/AI/FleshSentinel.cpp"
#include "scripts/BP/Actors/AI/FleshWorm.cpp"
#include "scripts/BP/Actors/AI/Grub.cpp"
#include "scripts/BP/Actors/AI/HiveDrone.cpp"
#include "scripts/BP/Actors/AI/HiveSoldier.cpp"
#include "scripts/BP/Actors/AI/HiveWorker.cpp"
#include "scripts/BP/Actors/AI/Leaper.cpp"
#include "scripts/BP/Actors/AI/LordoftheDead.cpp"
#include "scripts/BP/Actors/AI/LordoftheFlesh.cpp"
#include "scripts/BP/Actors/AI/Mite.cpp"
#include "scripts/BP/Actors/AI/MotherGrub.cpp"
#include "scripts/BP/Actors/AI/Nala.cpp"
#include "scripts/BP/Actors/AI/PrimagenDrone.cpp"
#include "scripts/BP/Actors/AI/PrimagenGuard.cpp"
#include "scripts/BP/Actors/AI/PrimagenTrooper.cpp"
#include "scripts/BP/Actors/AI/PurlinGunner.cpp"
#include "scripts/BP/Actors/AI/PurlinJuggernaut.cpp"
#include "scripts/BP/Actors/AI/PurlinWarClub.cpp"
#include "scripts/BP/Actors/AI/Raptoid.cpp"
#include "scripts/BP/Actors/AI/Raptor.cpp"
#include "scripts/BP/Actors/AI/SisterofDespair.cpp"
#include "scripts/BP/Actors/AI/SpiderHatchling.cpp"
#include "scripts/BP/Actors/AI/SwampWasp.cpp"

#include "scripts/BP/Actors/Maps/ScriptMap.cpp"

#include "scripts/BP/Server/ConsoleCommands.cpp"
#include "scripts/BP/Server/NetworkCommands.cpp"
#include "scripts/BP/Actors/Controllers/WorldChar.cpp"

void BP_Main()
{
	BP::Words::Load();
}
